;;; -*- Mode: Lisp; Syntax: Common-Lisp; Base: 10; Package: mcs -*-

;;           Copyright  1993    BY
;;           G M D  
;;           Schloss Birlinghoven
;;           D-53757 Sankt Augustin
;;           FRG

;;; -----------------------------------------------------------------------------------
;;;
;;; description:    This is the file of patches.
;;;
;;; notes:          
;;;
;;; contact:        Juergen Kopp, Harry Bretthauer
;;;
;;; history:
;;;          date:      author:             comments:
;;;          23.02.90   Harry Bretthauer    initial version


;;; -----------------------------------------------------------------------------------

(in-package "MCS")
 
;;; date       function                                           file
;;; author     comment

;;; -----------------------------------------------------------------------------------

;;; 23.06.93  compute-discriminating-fn                              gfn-core.lisp
;;;           Problem with compile in AKCL

(defun compute-discriminating-fn (gfn)
  (declare (optimize (speed 3) (safety 0)))
  (let ((methods (%%gfn-methods gfn))
        (signature (%%gfn-signature gfn)))
    (setf (%%gfn-discriminating-function gfn)
          (cond
           ((null methods) #'no-applicable-methods)
           ((equal signature '(t)) #'darg1-discriminator)
           ((equal signature '(() t)) #'arg1-darg2-discriminator)
           ((equal signature '(t ())) #'darg1-arg2-discriminator)
           ((equal signature '(t () ())) #'darg1-arg2-arg3-discriminator)
           ((equal signature '(t () () ())) #'darg1-arg2-arg3-arg4-discriminator)
           ((equal signature '(t t)) #'darg1-darg2-discriminator)
           ((equal signature '(t t ())) #'darg1-darg2-arg3-discriminator)
           ((equal signature '(t t () ())) #'darg1-darg2-arg3-arg4-discriminator)
           ((equal signature '(t t t)) #'darg1-darg2-darg3-discriminator)
           ((equal signature '(t t t ())) #'darg1-darg2-darg3-arg4-discriminator)
           ((equal signature '(())) #'arg1-discriminator)
           ((equal signature '(() ())) #'arg1-arg2-discriminator)
           ((equal signature '(() () ())) #'arg1-arg2-arg3-discriminator)

           (t #-:AKCL (compile () (compute-discriminating-lambda gfn methods))
	    #+:AKCL (error "Ask your MCS wizard to enlarge the predefined discriminators!!!")
	    )))))

(defun arg1-arg2-discriminator (gfn arg1 arg2)
  (declare (optimize (speed 3) (safety 0)))
  (or (cdr (mcs-assq () (%%gfn-combined-methods gfn)))
      (dynamic-lookup gfn (list arg1 arg2))))

(defun arg1-arg2-arg3-discriminator (gfn arg1 arg2 arg3)
  (declare (optimize (speed 3) (safety 0)))
  (or (cdr (mcs-assq () (%%gfn-combined-methods gfn)))
      (dynamic-lookup gfn (list arg1 arg2 arg3))))

(defun darg1-darg2-darg3-discriminator (gfn darg1 darg2 darg3)
  (declare (optimize (speed 3) (safety 0)))
  (or (cdr (mcs-assq (%class-of darg3)
		 (cdr (mcs-assq (%class-of darg2)
			      (cdr (mcs-assq (%class-of darg1)
					 (%%gfn-combined-methods gfn)))))))
      (dynamic-lookup gfn (list darg1 darg2 darg3))))

(defun darg1-darg2-darg3-arg4-discriminator (gfn darg1 darg2 darg3 arg4)
  (declare (optimize (speed 3) (safety 0)))
  (or (cdr (mcs-assq (%class-of darg3)
		 (cdr (mcs-assq (%class-of darg2)
			      (cdr (mcs-assq (%class-of darg1)
					 (%%gfn-combined-methods gfn)))))))
      (dynamic-lookup gfn (list darg1 darg2 darg3 arg4))))

;;; -----------------------------------------------------------------------------------

;;; 16.09.93   next-method-                                       system-m.lisp
;;; Bretthauer serious bug fixed!!!  
;;; Was: `%next-fns    instead of    `(cdr %next-fns)
;;; Although, this is a macro, recompiling of MCS is not required, only
;;; loading the patch and recompiling of the application.
;;; This bug happened because passing methods has been changed, 
;;; but not this macro. 

(defmacro next-method-p ()
  `(cdr %next-fns))

;;; -----------------------------------------------------------------------------------


;;; Unknown slot options in (DEFCLASS ...) forms are evaluated in newer versions of MCS.
;;; The frames in babylon require to quote them. This is a corresponding version
;;; of canonicalizing slot options:

#|
(defun compute-canonical-slot-specifications (slot-specs class-name)
  (declare (optimize (speed 3) (safety 0)))
  (let ((slot-names nil))
    (do ((rest-slot-specs slot-specs (rest rest-slot-specs))
         (name nil nil) (initform '<unbound> '<unbound>) (initfunction nil nil)
         (type nil nil) (documentation nil nil)
         (initarg nil nil) (defreader-forms nil) (defwriter-forms nil)
         (slot-class nil nil) (unknown-options nil nil) 
         (initargs nil) (result nil) (temp-result nil nil)) 
        ((null rest-slot-specs) 
         (list initargs defreader-forms defwriter-forms (reverse result)))
      (cond 
       ((consp (first rest-slot-specs))
        (setq name (caar rest-slot-specs))
        (doplist ((key value) (cdar rest-slot-specs))
          (case key
            (:reader (unless *during-bootstrap*
                       (push `(defreader ,value ,class-name ,name) defreader-forms)))
            (:writer (unless *during-bootstrap*
                       (push `(defwriter ,value ,class-name ,name) defwriter-forms)))
            (:accessor (unless *during-bootstrap*
                         (push `(defreader ,value ,class-name ,name) defreader-forms)
                         (push `(defwriter (setf ,value) ,class-name ,name) defwriter-forms)))
            (:initarg (if initarg
                        (error "For slot ~S of ~S :initarg option used twice."
                               name class-name)
                        (if (eq name value)
                          (setq initarg t
                                initargs (cons value initargs))
                          (error "Initarg ~S for slot ~S isn't eq to slot name."
                                 value name) )))
            (:allocation 
             (error "Don't use slot option :allocation for slot ~S any more." name))
            (:initform (if (eq initform '<unbound>)
                         (setq initform value
                               initfunction `#'(lambda () ,value))
                         (error "For slot ~S of ~S :initform option used twice."
                                name class-name)))
            (:type (if type 
                     (error "For slot ~S of ~S :type option used twice." name class-name)
                     (setq type `(and ,value))))
            (:slot-class (if slot-class
                           (error "For slot ~S of ~S :slot-class option used twice."
                                  name class-name)
                           (progn
                             (setq slot-class value)
                             (setf temp-result `('slot-class ',slot-class ,@temp-result)))))
            (:documentation (if documentation
                              (error "For slot ~S of ~S :documentation option used twice."
                                     name class-name)
                              (pe-store 'slot-documentation
                                        (list class-name name) value)))
            (t  (if (mcs-memq key *implemented-slot-options*)
                  (setq unknown-options
                          `(',key ',value ,@unknown-options))
                  (progn
                    (when *warn-if-unknown-slot-option*
                      (warn "Undefined slot option ~S for slot ~S in class ~S."
                            key name class-name))
                    (setq unknown-options
                          `(',key ',value ,@unknown-options))))) )))
       (t (setq name (first rest-slot-specs))))
      (if (mcs-memq name slot-names)
        (error "Slot ~S specified twice for class ~S." name class-name)
        (push name slot-names))
      
      (push
       `(list 'name ',name 'initform ',initform 'initfunction ,initfunction
              'type ',type 	;; 'readers ',readers 'writers ',writers 
              ,@unknown-options
              ,@temp-result)
       result))))
|#


;;; eof

